@@ -1,5 +1,5 @@
 // Copyright Contributors to the OpenImageIO project.
-// SPDX-License-Identifier: BSD-3-Clause and Apache-2.0
+// SPDX-License-Identifier: Apache-2.0
 // https://github.com/AcademySoftwareFoundation/OpenImageIO
 
 
@@ -9,7 +9,7 @@
 #include <memory>
 #include <stack>
 
-#include <boost/container/flat_set.hpp>
+#include <tsl/robin_set.h>
 
 #include <OpenImageIO/half.h>
 
@@ -82,6 +82,7 @@ public:
     bool skip_bad_frames = false;  // Just skip a bad frame, don't exit
     bool nostderr        = false;  // If true, use stdout for errors
     bool noerrexit       = false;  // Don't exit on error
+    bool create_dir      = false;
     std::string dumpdata_C_name;
     std::string full_command_line;
     std::string printinfo_metamatch;
@@ -126,6 +127,7 @@ public:
     bool output_dither;
     bool output_force_tiles;  // for debugging
     bool metadata_nosoftwareattrib;
+    bool metadata_history;
 
     // Options for --diff
     float diff_warnthresh;
@@ -139,8 +141,8 @@ public:
     ImageRecRef curimg;                    // current image
     std::vector<ImageRecRef> image_stack;  // stack of previous images
     std::map<std::string, ImageRecRef> image_labels;  // labeled images
-    ImageCache* imagecache = nullptr;                 // back ptr to ImageCache
-    ColorConfig colorconfig;                          // OCIO color config
+    std::shared_ptr<ImageCache> imagecache;           // back ptr to ImageCache
+    std::unique_ptr<ColorConfig> m_colorconfig;       // OCIO color config
     Timer total_runtime;
     // total_readtime is the amount of time for direct reads, and does not
     // count time spent inside ImageCache.
@@ -258,6 +260,17 @@ public:
         return r;
     }
 
+    void popbottom()
+    {
+        if (image_stack.size()) {
+            // There are images on the full stack -- get rid of the bottom
+            image_stack.erase(image_stack.begin());
+        } else {
+            // Nothing on the stack, so get rid of the current image
+            curimg = ImageRecRef();
+        }
+    }
+
     ImageRecRef top() { return curimg; }
 
     // How many images are on the stack?
@@ -358,6 +371,8 @@ public:
     // Merge stats from another Oiiotool
     void merge_stats(const Oiiotool& ot);
 
+    ColorConfig& colorconfig();
+
 private:
     CallbackFunction m_pending_callback;
     std::vector<const char*> m_pending_argv;
@@ -422,7 +437,7 @@ private:
 /// and potentially MIPmap levels for each subimage.
 class ImageRec {
 public:
-    ImageRec(const std::string& name, ImageCache* imagecache)
+    ImageRec(const std::string& name, std::shared_ptr<ImageCache> imagecache)
         : m_name(name)
         , m_imagecache(imagecache)
     {
@@ -459,7 +474,7 @@ public:
 
     // Initialize an ImageRec with the given spec.
     ImageRec(const std::string& name, const ImageSpec& spec,
-             ImageCache* imagecache);
+             std::shared_ptr<ImageCache> imagecache);
 
     ImageRec(const ImageRec& copy) = delete;  // Disallow copy ctr
 
@@ -580,15 +595,6 @@ public:
     }
     void clear_configspec() { m_configspec.reset(); }
 
-    /// Error reporting for ImageRec: call this with printf-like arguments.
-    /// Note however that this is fully typesafe!
-    template<typename... Args>
-    OIIO_DEPRECATED("Use errorfmt instead")
-    void errorf(const char* fmt, const Args&... args) const
-    {
-        append_error(Strutil::sprintf(fmt, args...));
-    }
-
     /// Error reporting for ImageRec: call this with printf-like arguments.
     /// Note however that this is fully typesafe!
     template<typename... Args>
@@ -615,7 +621,7 @@ private:
     std::vector<SubimageRec> m_subimages;
     std::time_t m_time;  //< Modification time of the input file
     TypeDesc m_input_dataformat;
-    ImageCache* m_imagecache = nullptr;
+    std::shared_ptr<ImageCache> m_imagecache;
     mutable std::string m_err;
     std::unique_ptr<ImageSpec> m_configspec;
 
@@ -776,13 +782,16 @@ public:
     // stack.
     OiiotoolOp(Oiiotool& ot, string_view opname, cspan<const char*> argv,
                int ninputs, setup_func_t setup_func = nullptr,
-               impl_func_t impl_func = nullptr)
+               impl_func_t impl_func          = nullptr,
+               ParamValueSpan control_options = {})
         : ot(ot)
         , m_nargs((int)argv.size())
         , m_nimages(ninputs + 1)
         , m_setup_func(setup_func)
         , m_impl_func(impl_func)
     {
+        m_control_options.assign(control_options.begin(),
+                                 control_options.end());
         if (Strutil::starts_with(opname, "--"))
             opname.remove_prefix(1);  // canonicalize to one dash
         m_opname = opname.substr(0, opname.find_first_of(':'));  // and no :
@@ -798,6 +807,12 @@ public:
         : OiiotoolOp(ot, opname, argv, ninputs, {}, impl_func)
     {
     }
+    OiiotoolOp(Oiiotool& ot, string_view opname, cspan<const char*> argv,
+               int ninputs, impl_func_t impl_func,
+               ParamValueSpan control_options)
+        : OiiotoolOp(ot, opname, argv, ninputs, {}, impl_func, control_options)
+    {
+    }
     virtual ~OiiotoolOp() {}
 
     // The operator(), function-call mode, does most of the work. Although
@@ -886,10 +901,12 @@ public:
             m_img.resize(nimages());
             for (int m = 0, nmip = ir(0)->miplevels(); m < nmip; ++m) {
                 m_current_miplevel = m;
-                for (int i = 0; i < nimages(); ++i)
-                    m_img[i] = &((*ir(i))(std::min(s, ir(i)->subimages() - 1),
-                                          std::min(m, ir(i)->miplevels(s))));
-
+                for (int i = 0; i < nimages(); ++i) {
+                    // If the subimage doesn't exist, just use the last
+                    int ss   = std::min(s, ir(i)->subimages() - 1);
+                    m_img[i] = &(
+                        (*ir(i))(ss, std::min(m, ir(i)->miplevels(ss))));
+                }
                 if (subimage_is_active(s)) {
                     // Call the impl kernel for this subimage
                     bool ok = impl(m_img);
@@ -1005,7 +1022,35 @@ public:
             }
         }
         all_subimages |= m_options.get_int("allsubimages", ot.allsubimages);
-        return all_subimages ? (nimages() > 1 ? ir(1)->subimages() : 1) : 1;
+
+        // How many subimages are we going to operate on? There are a few
+        // strategies for handling the decision when the input images differ
+        // in the number of subimages. The default is to operate on subimage
+        // list from the first image input.
+        int n = 1;
+        if (all_subimages) {
+            OIIO_DASSERT(nimages() >= 1);
+            std::string subimage_policy = m_control_options["subimage_policy"];
+            if (subimage_policy == "max") {
+                // operate on the maximum number of subimages of the inputs
+                for (int i = 1; i < nimages(); ++i)
+                    n = std::max(n, ir(i)->subimages());
+            } else if (subimage_policy == "min") {
+                // operate on the minimum number of subimages of the inputs
+                if (nimages() > 1)
+                    n = ir(1)->subimages();
+                for (int i = 2; i < nimages(); ++i)
+                    n = std::min(n, ir(i)->subimages());
+            } else if (subimage_policy == "last") {
+                // operate on the subimages of the last input
+                n = ir(nimages() - 1)->subimages();
+            } else /* if (subimage_policy == "first")*/ {  // default
+                // default: operate on the subimages of the first input
+                if (nimages() > 1)
+                    n = ir(1)->subimages();
+            }
+        }
+        return n;
     }
 
     // Is the given subimage in the active set to be operated on by this op?
@@ -1039,7 +1084,10 @@ public:
     }
 
     int nargs() const { return m_nargs; }
-    string_view args(int i) const { return m_args[i]; }
+    string_view args(int i) const
+    {
+        return i < m_nargs ? m_args[i] : string_view();
+    }
     int nimages() const { return m_nimages; }
     string_view opname() const { return m_opname; }
     ParamValueList& options() { return m_options; }
@@ -1096,7 +1144,7 @@ protected:
     std::vector<ImageBuf*> m_img;
     std::vector<string_view> m_args;
     ParamValueList m_options;
-    typedef boost::container::flat_set<int> FastIntSet;
+    typedef tsl::robin_set<int> FastIntSet;
     FastIntSet subimage_includes;  // Subimages to operate on (empty == all)
     FastIntSet subimage_excludes;  // Subimages to skip for the op
     setup_func_t m_setup_func;
@@ -1104,6 +1152,7 @@ protected:
     new_output_imagerec_func_t m_new_output_imagerec_func;
     int m_current_subimage;  // for impl(), which subimage are we on?
     int m_current_miplevel;  // for impl(), which miplevel are we on?
+    ParamValueList m_control_options;
 };
 
 
